#!/usr/bin/env perl
# @(#) pg_trompe application to replicate database events between two databases
# $Header: /code/pg_trompe/replication.pl,v 1.37 2009/02/20 21:08:32 doj Exp $

=head1 NAME

replication.pl - pg_trompe application to replicate database events between two databases

=cut

use DBD::Pg v2.5.0 qw(:async);
use DBI;
use Getopt::Long;
use Sys::Syslog;
use strict;

use constant PRIMARY_KEY => 'nr';
use constant LO_KEY => 'lo';

my $sth;
my $sth2;
my @after_replication;		# contains references to subs
my $exitval=0;

=head1 SYNOPSIS

replication.pl --local_dbname <databasename> [--local_dbuser <username>] [--local_dbpass <passwort>] [--local_dbhost <hostname>] [--local_dbport <port>] [--remote_dbname <databasename>] [--remote_dbuser <username>] [--remote_dbpass <passwort>] [--remote_dbhost <hostname>] [--remote_dbport <port>] [--syslog_fac <facility string>] [--compare_db] [--stats] [--verbose]

=head1 OPTIONS

=over

=item local_dbname

Required option. Sets the local dbname.

=item local_dbuser

Username to use for the local db.

=item local_dbpass

Password to use for the local db.

=item local_dbhost

Hostname of local db.

=item local_dbport

TCP Port of local db.

=item remote_dbname

The remote dbname. If not set the name of local_dbname is used.

=item remote_dbuser

Username for the remote db.

=item remote_dbpass

Password for the remote db.

=item remote_dbhost

Hostname of the remote db.

=item remote_dbport

Password of the remote db.

=item syslog_fac

Syslog facility to use for reporting information and errors. Defaults
to LOCAL6. If set to the empty string, syslog will not be used. Any
string which is supported by Sys::Syslog can be used.

=item compare_db

If set all tables in replicated databases are compared after the replication run.

=item stats

If set number of outstanding replication events are shown for both databases.

=item verbose

If set to a true value, output all SQL commands in replication.

=back

=cut

my $local_dbname;
my $local_dbuser;
my $local_dbpass;
my $local_dbhost;
my $local_dbport;
my $remote_dbname;
my $remote_dbuser;
my $remote_dbpass;
my $remote_dbhost;
my $remote_dbport;
my $syslog_fac="local6";
my $compare_db;
my $stats;
my $verbose;

$_=GetOptions(
	      "local_dbname=s"	=> \$local_dbname,
	      "local_dbuser=s"	=> \$local_dbuser,
	      "local_dbpass=s"	=> \$local_dbpass,
	      "local_dbhost=s"	=> \$local_dbhost,
	      "local_dbport=i"	=> \$local_dbport,
	      "remote_dbname=s"	=> \$remote_dbname,
	      "remote_dbuser=s"	=> \$remote_dbuser,
	      "remote_dbpass=s"	=> \$remote_dbpass,
	      "remote_dbhost=s"	=> \$remote_dbhost,
	      "remote_dbport=i"	=> \$remote_dbport,
	      "syslog_facility=s" => \$syslog_fac,
	      "compare_db" => \$compare_db,
	      "stats" => \$stats,
	      "verbose" => \$verbose,
	     );

openlog("pg_trompe", "pid", $syslog_fac) or die "could not open syslog";
$SIG{'__DIE__'} = sub {
    my $e="DIE: ".join('; ', @_);
    eval {
	syslog('crit', "$e\n") if $syslog_fac;
    };
    print STDERR isodate()." $e\n";
};
$SIG{'__WARN__'} = sub {
    my $w="WARN: ".join('; ', @_);
    eval {
	syslog('warning', "$w\n") if $syslog_fac;
    };
    print STDERR isodate()." $w\n";
};
sub mylog
{
    my $l=join(' ', @_);
    chomp $l;
    eval {
	syslog('notice', "$l\n") if $syslog_fac;
    };
    print STDERR isodate()." $l\n";
}
sub isodate
{
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    return sprintf("%04i-%02i-%02iT%02i:%02i:%02i", $year+1900, $mon+1, $mday, $hour, $min, $sec);
}

# connect to databases

sub connectdb
{
    my ($dbname, $dbuser, $dbpass, $dbhost, $dbport) = @_;
    my $c="dbi:Pg:dbname=$dbname".($dbhost?";host=$dbhost":'').($dbport?";port=$dbport":'');
    mylog("connect to $c $dbuser $dbpass");
    my $dbh=DBI->connect($c, $dbuser, $dbpass, { RaiseError => 1, AutoCommit => 0 }) or die "could not connect to $c";
    return begin_transaction($dbh);
}

sub begin_transaction
{
    my ($dbh) = @_;
    return unless ref $dbh;
    my $q="SET CONSTRAINTS ALL DEFERRED";
    $dbh->do($q) or die "could not set $q";
    $q="SET client_encoding = 'UTF-8'";
    $dbh->do($q) or die "could not set $q";
    $q="SET xmloption TO CONTENT";
    $dbh->do($q) or die "could not set $q";
    return $dbh;
}

die "local_dbname not set" unless $local_dbname;
my $ldbh=connectdb($local_dbname, $local_dbuser, $local_dbpass, $local_dbhost, $local_dbport);
showstats($ldbh, "Local Database: $local_dbname") if $stats;

$remote_dbname ||= $local_dbname;
my $rdbh=connectdb($remote_dbname, $remote_dbuser, $remote_dbpass, $remote_dbhost, $remote_dbport);
showstats($rdbh, "Remote Database: $remote_dbname") if $stats;

goto replication_ende if $stats;

# get node numbers
my $q="select val from replication.config where key = 'node_num'";
my ($local_node) = $ldbh->selectrow_array($q);
die "replication not configured on local" unless defined $local_node;
my ($remote_node) = $rdbh->selectrow_array($q);
die "replication not configured on remote" unless defined $remote_node;
die "both nodes have the same node num ($remote_node)" if $local_node == $remote_node;

# check that replication.tables are equal
my %tables;

my $q="select * from replication.tables order by name";
$sth=$ldbh->prepare($q); $sth->execute();
$sth2=$rdbh->prepare($q); $sth2->execute();
while(my $h=$sth->fetchrow_hashref())
{
    my $h2=$sth2->fetchrow_hashref() or die "remote replication.tables is short";
    die "replication.tables are not equal" if $h->{name} ne $h2->{name};
    die "replication.tables $h->{name} lo setting is not equal" if $h->{lo} != $h2->{lo};

    # check that tables are equal
    my $tablefullname=$h->{name};
    my $schema="public";
    my $tablename=$tablefullname;
    if($tablename =~ /^(\w+)\.(\w+)$/)
    {
	$schema=$1;
	$tablename=$2;
    }
    mylog("check $schema.$tablename") if $verbose;

    $tables{$tablefullname} = { PRIMARY_KEY() => PRIMARY_KEY() };

    my $sth3=$ldbh->column_info(undef, $schema, $tablename, '%') or die "could not get local column info";
    while(my $h3=$sth3->fetchrow_hashref())
    {
	$tables{$tablefullname}->{$h3->{COLUMN_NAME}} = $h3->{TYPE_NAME};
	mylog("  local $tablefullname $h3->{COLUMN_NAME} $h3->{TYPE_NAME}") if $verbose;
    }
    $sth3->finish();

    $sth3=$rdbh->column_info(undef, $schema, $tablename, '%') or die "could not get remote column info";
    while(my $h3=$sth3->fetchrow_hashref())
    {
	die "local database $tablefullname does not have $h3->{COLUMN_NAME} column" unless exists $tables{$tablefullname}->{$h3->{COLUMN_NAME}};
	die "$tablefullname column $h3->{COLUMN_NAME} type not equal" if $tables{$tablefullname}->{$h3->{COLUMN_NAME}} ne $h3->{TYPE_NAME};
	mylog("  remote $tablefullname $h3->{COLUMN_NAME} $h3->{TYPE_NAME}") if $verbose;

	# TODO: local database table might contain columns not in remote table
    }
    $sth3->finish();
}
$sth->finish();
$sth2->finish();

# get node description
$q="select val from replication.config where key = 'node_desc'";
my ($local_desc) = $ldbh->selectrow_array($q);
my ($remote_desc) = $rdbh->selectrow_array($q);

# check that nodes are configured to sync each other
mylog("check node registring");

sub check_node
{
    my ($dbh, $node_num, $desc) = @_;
    return unless ref $dbh;
    return unless $node_num;
    my ($c) = $dbh->selectrow_array("select count(*) from replication.nodes where num=?", undef, $node_num);
    if($c == 0)
    {
	mylog("registering Node $node_num $desc");
	$dbh->do("insert into replication.nodes(num,description) values(?,?)", undef, $node_num, $desc) or die "could not register node $node_num";
	push(@after_replication, sub {
		 mylog("Set all Sequences on Node $node_num");
		 $dbh->do("select replication.setallseq()");
	     } );
    }
}
check_node($ldbh, $remote_node, ($remote_desc || '_nodesc_'));
check_node($rdbh, $local_node, ($local_desc || '_nodesc_'));

sub replicate
{
    my ($from, $from_num, $to, $to_num) = @_;
    return unless ref $from;
    return unless ref $to;
    return unless defined $from_num;
    return unless defined $to_num;

    my $from_md5_sth;

    mylog("replicate $from_num -> $to_num");

    my $fn="/tmp/$$.lo";
    END { unlink $fn if -f $fn }

    my ($r) = $to->selectrow_array("select replication.start(?)", undef, $from_num);
    die "could not start replication" unless $r == 0;

    my $sth=$from->prepare("select * from replication.log where node=? order by id");
    $sth->execute($to_num);
    my $Z = $sth->rows();
    my $z=0;
    while(my $log=$sth->fetchrow_hashref())
    {
	my $l="$from_num->$to_num ".(++$z)."/$Z ($log->{id}) ";

	my $primary_key=$log->{PRIMARY_KEY()};

	if($log->{action} eq 'INSERT')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()});

	    my $q="select ".select_cols_str($log->{tablename})." from $log->{tablename} where ".PRIMARY_KEY."=?";
	    mylog("  $q") if $verbose;
	    my $h=$from->selectrow_hashref($q, undef, $primary_key);
	    unless(defined $h->{PRIMARY_KEY()})
	    {
		mylog("  now rows found.");
		next;
	    }

	    if($h->{LO_KEY()})
	    {
		my ($comment) = $from->selectrow_array("select obj_description(?)", undef, $h->{LO_KEY()});

		$from->func($h->{LO_KEY()}, $fn, 'lo_export') or die "could not lo_export $log->{tablename}::$primary_key ".$h->{LO_KEY()};
		mylog("  ".(-s $fn)." bytes");
		$h->{LO_KEY()} = $to->func($fn, 'lo_import') or die "could not lo_import $fn";

		$to->do("comment on large object ".$h->{LO_KEY()}." is ?", undef, $comment) if $comment;
	    }

	    my @q;		# collects the column values
	    my $v='';		# collects string in VALUES()
	    $q='';		# collects string in tablename()
	    while(my ($col,$val) = each(%$h))
	    {
		$v.=',' if length($v);
		$q.=',' if length($q);

		$q.=$col;

		if($tables{$log->{tablename}}->{$col} eq 'xml')
		{
		    $v.="XMLPARSE(CONTENT ?)";
		}
		else
		{
		    $v.='?';
		}

		push(@q, $val);
	    }
	    $q="insert into $log->{tablename}($q) values($v)";
	    mylog("  $q := ".join(',', @q)) if $verbose;
	    $to->do($q, undef, @q) or die "could not insert";
	}
	elsif($log->{action} eq 'UPDATE')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()});

	    my $q="select ".select_cols_str($log->{tablename})." from $log->{tablename} where ".PRIMARY_KEY."=?";
	    mylog("  $q") if $verbose;
	    my $h=$from->selectrow_hashref($q, undef, $primary_key);
	    unless(defined $h->{PRIMARY_KEY()})
	    {
		mylog("  now rows found.");
		next;
	    }

	    my @q;
	    $q='';
	    while(my ($col,$val) = each(%$h))
	    {
		next if $col eq PRIMARY_KEY || $col eq LO_KEY;
		$q.=", " if length($q);

		$q.="$col=";
		if($tables{$log->{tablename}}->{$col} eq 'xml')
		{
		    $q.='XMLPARSE(CONTENT ?)';
		}
		else
		{
		    $q.='?';
		}
		push(@q, $val);
	    }
	    $q="update $log->{tablename} set $q where ".PRIMARY_KEY."=?";
	    push(@q, $primary_key);
	    mylog("  $q := ".join(',', @q)) if $verbose;
	    $to->do($q, undef, @q) or die "could not update";

	    my $from_oid=$h->{LO_KEY()};
	    if($from_oid)
	    {
		# start first query async
		$from_md5_sth ||= $from->prepare("select md5(?::oid), obj_description(?)", {pg_async => PG_ASYNC}) or die "could not prepare from_md5_sth";
		$from_md5_sth->execute($from_oid, $from_oid);
		# second query in sync
		my ($to_oid, $to_md5, $to_comment) = $to->selectrow_array("select ".LO_KEY.", md5(".LO_KEY."::oid), obj_description(".LO_KEY.") from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key);
		# get result of first query
		$from->pg_result;
		my ($from_md5, $from_comment) = $from_md5_sth->fetchrow_array();
		$from_md5_sth->finish();

		if($from_md5 ne $to_md5)
		{
		    $from->func($from_oid, $fn, 'lo_export') or die "could not lo_export $log->{tablename}::$primary_key $from_oid";
		    mylog("  ".(-s $fn)." bytes");
		    if($to_oid)
		    {
			# TODO: rsync
			mylog("  overwrite target lo");
			my $fd=$to->func($to_oid, $to->{pg_INV_WRITE}, 'lo_open');
			die "could not open $log->{tablename}::$primary_key $to_oid" unless defined $fd;
			open(A, $fn) or die "could not open $fn";
			my $lo_len=0;
			my $buf;
			while(sysread(A, $buf, 4096))
			{
			    my $buflen=length($buf);
			    $lo_len+=$buflen;
			    my $w=$to->func($fd, $buf, $buflen, 'lo_write');
			    # TODO: check that $w == $buflen. Develop the usual loop around lo_write to write all bytes in buf.
			    die "could not lo_write complete buffer $w != $buflen" if $w != $buflen;
			}
			close(A);
			$to->func($fd, $lo_len, 'lo_truncate');
			$to->func($fd, 'lo_close');
		    }
		    else
		    {
			mylog("  create new target lo");
			$to_oid = $to->func($fn, 'lo_import') or die "could not lo_import $fn";
			$to->do("update $log->{tablename} set ".LO_KEY."=? where ".PRIMARY_KEY."=?", undef, $to_oid, $primary_key) or die "could not update lo oid in target";
		    }
		}
		else
		{
		    mylog("  large objects are equal");
		}

		mylog("  from_comment=$from_comment to_comment=$to_comment");
		if($from_comment ne $to_comment)
		{
		    $to->do("comment on large object $to_oid is ?", undef, $from_comment);
		}
	    }
	}
	elsif($log->{action} eq 'DELETE')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()});

	    $to->do("delete from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key) or die "could not delete from $log->{tablename} #$primary_key";
	}
	elsif($log->{action} eq 'DDL')
	{
	    mylog("$l DDL: $log->{tablename}");
	    $to->do("select replication.ddl(?)", undef, $log->{'tablename'}) or die "could not execute DDL";
	}
	else
	{
	    warn "$l unknown action $log->{action}";
	}

	$from->do("delete from replication.log where id=?", undef, $log->{id}) or die "could not delete log #$log->{id}";
    }
    $sth->finish();

    ($r) = $to->selectrow_array("select replication.stop(?)", undef, $from_num);
    die "could not stop replication" unless $r == 0;

    $from->do("update replication.nodes set lastsync=now() where num=?", undef, $to_num) or die "could not update lastsync on from Node #$from_num";

    mylog("commit");
    $to->commit() or die "could not commit on destination node #$to_num";
    $from->commit() or die "could not commit on source node #$from_num";

    begin_transaction($from);
    begin_transaction($to);

    mylog("replicate $from_num -> $to_num finished");
}

eval
{
    replicate($ldbh,$local_node,$rdbh,$remote_node);
    replicate($rdbh,$remote_node,$ldbh,$local_node);
    if($#after_replication >= 0)
    {
	mylog("running after_replication hooks");
	foreach my $c (@after_replication)
	{
	    $c->() if ref($c) eq 'CODE';
	}
	$ldbh->commit();
	$rdbh->commit();
    }
};
if($@)
{
    warn "Transaction aborted because $@";
    eval { $ldbh->rollback };
    eval { $rdbh->rollback };
    mylog("exit failure");
    exit 1;
}

if($compare_db)
{
    mylog("comparing databases");
    while(my ($tablename, $tab) = each(%tables))
    {
	my $m='nr::text';
	while(my ($col,$typ) = each(%$tab))
	{
	    my $type=get_col_type($ldbh, $tablename, $col);
	    die "column $col of $tablename has no identical type" if $type ne get_col_type($rdbh, $tablename, $col);

	    $col="array_to_string($col, ',')" if $type =~ /\[\]$/; # array
	    if($type eq 'bit')
	    {
		$col="bit_send($col)";
		$type='bytea';
	    }
	    if($type eq 'bit varying')
	    {
		$col="varbit_send($col)";
		$type='bytea';
	    }
	    if($type eq 'circle')
	    {
		$col="circle_send($col)";
		$type='bytea';
	    }
	    if($type eq 'point')
	    {
		$col="point_send($col)";
		$type='bytea';
	    }
	    if($type eq 'line')
	    {
		$col="line_send($col)";
		$type='bytea';
	    }
	    if($type eq 'lseg')
	    {
		$col="lseg_send($col)";
		$type='bytea';
	    }
	    if($type eq 'box')
	    {
		$col="box_send($col)";
		$type='bytea';
	    }
	    if($type eq 'polygon')
	    {
		$col="poly_send($col)";
		$type='bytea';
	    }
	    if($type eq 'path')
	    {
		$col="path_send($col)";
		$type='bytea';
	    }
	    if($type eq 'xml')
	    {
		$col="XMLSERIALIZE(CONTENT $col AS text)";
	    }
	    $col="${col}::int" if $type eq 'boolean';
	    $col="cash_words($col)" if $type eq 'money';
	    $col="md5($col)" if $type eq 'oid' || $type eq 'bytea';
	    $m.=" || COALESCE(${col}::text, 'nil')";
	}

	my $q="select nr, md5($m) as md5 from $tablename order by nr";

	mylog(" compare $tablename with: $q");

	# start the first query async
	my $sthl=$ldbh->prepare($q, {pg_async => PG_ASYNC}) or die "could not prepare compare local";
	$sthl->execute();

	# second query in sync
	my $sthr=$rdbh->prepare($q) or die "could not prepare compare remote";
	$sthr->execute();

	# wait for first query
	$ldbh->pg_result();

	my $l={ nr => -1, md5 => 'a' };
	my $r={ nr => -1, md5 => 'a' };
	while($l && $r)
	{
	    if($l->{nr} < $r->{nr})
	    {
		mylog(" ${tablename}::$l->{nr} is missing on remote");
		$l=$sthl->fetchrow_hashref();
		$exitval=2;
	    }
	    elsif($l->{nr} > $r->{nr})
	    {
		mylog(" ${tablename}::$r->{nr} is missing on local");
		$r=$sthr->fetchrow_hashref();
		$exitval=3;
	    }
	    else
	    {
		if($l->{md5} ne $r->{md5})
		{
		    mylog(" ${tablename}::$r->{nr} is not equal");
		    $exitval=4;
		}
		$l=$sthl->fetchrow_hashref();
		$r=$sthr->fetchrow_hashref();
	    }
	}
	$sthl->finish();
	$sthr->finish();
    }
}

replication_ende:
$ldbh->disconnect();
$rdbh->disconnect();
$exitval ? mylog("compare failure") : mylog("exit success");
exit $exitval;

sub get_col_type
{
    my ($dbh, $tablename, $colname) = @_;
    return unless ref $dbh;
    return unless $tablename;
    return unless $colname;
    return unless $tablename =~ /(.+)\.(.+)/;
    my $schemaname=$1;
    $tablename=$2;
    my $sth = $dbh->column_info(undef, $schemaname, $tablename, $colname);
    my $h=$sth->fetchrow_hashref();
    $sth->finish();
    return $h->{'TYPE_NAME'};
}

sub select_cols_str
{
    my ($tablename) = @_;
    return '*' unless $tablename;
    return '*' unless exists $tables{$tablename};
    return '*' unless ref $tables{$tablename};
    my $s='';
    while(my ($col, $type) = each %{$tables{$tablename}})
    {
	$s.=", " if $s;
	if($type eq 'xml')
	{
	    $s.="XMLSERIALIZE(CONTENT $col as text) as $col";
	}
	else
	{
	    $s.=$col;
	}
    }
    return $s;
}

sub showstats
{
    my ($dbh, $n) = @_;
    return unless ref $dbh;

    print "$n\n";
    my $a = $dbh->selectall_arrayref("select n.description, r.tablename, count(*)
from replication.log as r left outer join replication.nodes as n on (r.node = n.num)
group by n.description, r.tablename");

    foreach my $r (@$a)
    {
	print join("\t", @$r)."\n";
    }

    print "\n";
}

__END__

=head1 RETURN VALUE

=over

=item 0 success

=item 1 replication failed

=item 2 a row is missing on the remote db. Resulting from -compare_db

=item 3 a row is missing on the local db. Resulting from -compare_db

=item 4 a row is not equal on local and remote. Resulting from -compare_db

=back

=head1 USAGE

=head2 Database Schema Requirements

=over

=item (pseudo) primary key

Each replicated table must have a column of the following syntax: C<nr
serial8 primary key> or C<nr serial8 not null unique>. Once those
C<nr> values are assigned with the initial C<INSERT> statement they
may not be altered.

=item large object support

If the replication should handle large objects, they must be managed
by a replicated table which also contains a column of the following
format: C<lo oid unique>. Thus each row of this table can
manage/contain at most one large object. The lo column may be C<NULL>.

=item unique values

To prevent I<uniqueness conflicts> the tables should not contain
C<unique> columns. If you need unique values you should either manage
them at the application layer or be really careful to not insert two
unique values in two databases at the same time. Otherwise the
replication might fail and the databases need to be manually cleaned
(the uniqueness conflict resolved) before replication can continue.

=back

=head2 Replication Setup

Before replication data between databases they must be configured. The
replicated tables may contain data before you run the steps described
below and this data will be included in the initial replication.

=over

=item Execute F<schema.sql>, F<functions.sql>, F<tables.sql> in each
database. This creates the C<replication> schema which contains all
functions and tables used by pg_trompe.

=item Set the Node number of the database with C<select
replication.node_num(1, 'Node Description');> Replace the C<1> with
the node number. Each node needs to have a unique number. By default
the numbers 0-9 are supported. To raise this limit see section L</"More
than 10 Nodes">.

=item For replicated table do C<select
replication.register('tablename');> or C<select
replication.registerlo('tablename');> depending if the table manages
large objects (and thus contains an C<lo> column) or not.

=back

Your database is now prepared for replication.

=head2 Drop replication

If you want to stop replication a database follow these steps.

=over

=item Tell all I<neighbour> databases that the node goes out of
replication by executing C<select replication.drop_node(node num);> on
them. C<node num> is the node number of the to-be dropped node.

=item Execute F<drop.sql> on the dropped node to remove the C<replication> schema.

=back

=head2 Replication

To replicate between two nodes use the F<replication.pl> program. You
have to supply database names and possibly usernames, password,
hostnames and TCP ports so the application knows how to connect to the
two databases involved. Upon the first replication run the two
databases are I<introduced> to each other and their table contents are
replicated. Afterwards every change on the replicated tables is
tracked in the C<replication.log>

=head2 TRUNCATE support

The command C<truncate> currently does not raise triggers, so it can't
be monitored by external applications. If you want to truncate
replicated tables use C<select replication.truncate('<ablename')>
instead.

=head2 DDL support

You can issue DDL command on replicated tables by using the function
C<replication.dll>. As parameter you supply one string which contains
the DDL command. On the local node the command is executed immediately
and it is replicated like any other events to remote databases. This
means that the order of DDL and non-DDL command is preserved. Example:
C<select replication.ddl('alter table user alter name set not null');>

=head2 More than 10 Nodes

By default pg_trompe supports up to 10 nodes. If you need more nodes
you have to change the configuration I<variable> C<maxnodes> in the
C<replication.config> table before registering any tables. This has to
be done on all databases and of course the value of C<maxnodes> needs
to be the same on all databases. Example: C<update replication.config
set maxnodes=100;>

=head1 CONSTANTS

The following perl constants influence the applications setup.

=head2 PRIMARY_KEY

The column name of (pseudo) primary keys.

=head2 LO_KEY

The column name of large object identifiers.

=head1 AUTHOR

Dirk Jagdmann E<lt>doj@cubic.orgE<gt>

Homepage S<http://llg.cubic.org/pg_trompe>

=cut
